import base64
from io import StringIO
from typing import List, Optional, Dict

from playhouse.shortcuts import model_to_dict
from pydantic import BaseModel

from lemon_rag.api.base import handle_request_with_pydantic, add_route, handle_chat_auth
from lemon_rag.api.local import get_user
from lemon_rag.configs.api_config import runtime_config
from lemon_rag.dependencies.data_access import data_access, KnowledgeBasePermission
from lemon_rag.lemon_runtime import models, wrappers
from lemon_rag.llm.agents.rag.retrieval.retrive_knowledgebase import retrieve_related_paragraphs, RelatedSentence
from lemon_rag.llm.agents.rag.vectorization.content_extractor import TxtExtractor
from lemon_rag.llm.agents.rag.vectorization.runner import start_parse_document
from lemon_rag.protocols.chat import ChatRole, Session, KnowledgeBase
from lemon_rag.utils import log
from lemon_rag.utils.file_utils import get_file_hash
from lemon_rag.utils.response_utils import response, ErrorCodes


class ListSessionRequest(BaseModel):
    version: int


class ListSessionResponse(BaseModel):
    up_to_date: bool = False
    sessions: Optional[List[Session]] = None


@handle_chat_auth
@handle_request_with_pydantic(ListSessionRequest)
def list_session(req: ListSessionRequest):
    assistant_session, _ = data_access.get_or_create_session(get_user(), ChatRole.assistant)
    notification_session, _ = data_access.get_or_create_session(get_user(), ChatRole.notification_center)
    return response(data=ListSessionResponse(sessions=[
        Session(**model_to_dict(assistant_session)),
        Session(**model_to_dict(notification_session))
    ]))


class GetNotificationCountRequest(BaseModel):
    version: int


class GetNotificationCountResponse(BaseModel):
    version: int
    unread_count: int


@handle_chat_auth
@handle_request_with_pydantic(GetNotificationCountRequest)
def get_notification_count(req: GetNotificationCountRequest):
    notification_session, _ = data_access.get_or_create_session(get_user(), ChatRole.notification_center)
    sync_history: models.SyncHistoryTab = notification_session.sync_history.get()
    return response(data=GetNotificationCountResponse(
        version=notification_session.version,
        unread_count=(notification_session.last_msg_id or 0) - (sync_history.last_read_id or 0)
    ))


class ReadNotificationsRequest(BaseModel):
    msg_id: int


@handle_chat_auth
@handle_request_with_pydantic(ReadNotificationsRequest)
def read_notifications(req: ReadNotificationsRequest):
    notification_session, _ = data_access.get_or_create_session(get_user(), ChatRole.notification_center)
    data_access.read_message(notification_session, msg_id=req.msg_id)
    return response()


class UploadFileRequest(BaseModel):
    file_content: str
    filename: str
    knowledge_base_id: int


class UploadFilesResponse(BaseModel):
    id: int
    filename: str
    origin_filename: str


def extract_extension(filename: str) -> str:
    return filename.split('.')[-1]


@handle_chat_auth
@handle_request_with_pydantic(UploadFileRequest)
def upload_knowledge_base_file(req: UploadFileRequest):
    file_content: bytes = base64.b64decode(req.file_content)
    file_size: int = len(file_content)
    extension = extract_extension(req.filename)
    # check file basic
    if file_size <= 0:
        return response(code=ErrorCodes.empty_file)
    if file_size >= runtime_config.max_file_bytes:
        return response(code=ErrorCodes.file_size_exceeded)
    if extension not in runtime_config.supported_file_types:
        return response(code=ErrorCodes.file_extension_invalid)

    # check kb existence
    kb = data_access.get_knowledge_base_by_id(req.knowledge_base_id)
    if not kb:
        log.info("knowledge base not found, id=%s", req.knowledge_base_id)
        return response(code=ErrorCodes.knowledge_base_not_found)

    # check access
    access = data_access.get_knowledge_base_access(get_user(), kb)
    if not access:
        log.info("no access found for the knowledge_base=%s", kb)
        return response(code=ErrorCodes.permission_denied)

    if not (access.permission & KnowledgeBasePermission.write):
        log.info(
            "the user has permission=%s, but permission=%s required",
            access.permission,
            KnowledgeBasePermission.write
        )
        return response(code=ErrorCodes.permission_denied)

    #
    if data_access.get_knowledge_base_file_count(kb) >= kb.max_files:
        return response(code=ErrorCodes.file_count_exceeded)

    # check hash
    file_hash = get_file_hash(file_content)
    if file := data_access.upload_file_by_hash(get_user(), file_hash):
        return response(data=UploadFilesResponse(**model_to_dict(file)))

    oss_file = wrappers.file_system.create_file(file_content, file_hash)
    oss_file.save()
    log.info("upload file, filename=%s, uploaded path=%s", req.filename, oss_file.url)
    file = data_access.create_file(oss_file.url, extension, req.filename, file_size, file_hash, kb, get_user())

    #
    buf = StringIO(file_content.decode("utf-8"))
    paragraphs = TxtExtractor().extract_content(buf)
    log.info("extracted paragraphs count=%s from %s", len(paragraphs), req.filename)
    start_parse_document(file, paragraphs)
    return response(data=UploadFilesResponse(**model_to_dict(file)))


class ListKnowledgeBasesRequest(BaseModel):
    version: int


class ListKnowledgeBasesResponse(BaseModel):
    knowledge_bases: List[KnowledgeBase]


@handle_chat_auth
@handle_request_with_pydantic
def list_knowledge_bases(req: ListKnowledgeBasesRequest):
    return response()


class SendMessageRequest(BaseModel):
    text: str
    session_id: int


@handle_chat_auth
@handle_request_with_pydantic
def send_message(request: SendMessageRequest):
    return response()


class RateMessageRequest(BaseModel):
    msg_id: int


@handle_chat_auth
@handle_request_with_pydantic
def rate_message(abc):
    response()


class RetrySendMessageRequest(BaseModel):
    msg_id: int


@handle_chat_auth
@handle_request_with_pydantic
def retry_send_message():
    # 判断是否是最新一条消息，如果是则复用发送消息逻辑
    pass


class SearchKnowledgebaseRequest(BaseModel):
    query: str
    knowledgebase_id: int


@handle_chat_auth
@handle_request_with_pydantic
def search_knowledgebase(request: SearchKnowledgebaseRequest):
    knowledgebase = data_access.get_knowledge_base_by_id(request.knowledgebase_id)
    if not knowledgebase:
        return response(code=ErrorCodes.knowledge_base_not_found)
    file_ids = [file.id for file in knowledgebase.files]
    res: Dict[str, RelatedSentence] = retrieve_related_paragraphs(request.query, file_ids)
    return response(data=list(res.values()))


add_route("list_session", list_session)
add_route("get_notification_count", get_notification_count)
add_route("read_notifications", read_notifications)
add_route("upload_knowledge_base_file", upload_knowledge_base_file)
add_route("send_message", send_message)
add_route("rate_message", rate_message)
add_route("retry_send_message", retry_send_message)
add_route("list_knowledge_bases", list_knowledge_bases)
add_route("search_knowledgebase", search_knowledgebase)



from typing import Optional, Union, Any
from uuid import UUID

from langchain_core.callbacks import BaseCallbackHandler
from langchain_core.outputs import GenerationChunk, ChatGenerationChunk


class TextCallbackHandler(BaseCallbackHandler):
    def __init__(self):
        self.value = ""

    def on_llm_new_token(
            self,
            token: str,
            *,
            chunk: Optional[Union[GenerationChunk, ChatGenerationChunk]] = None,
            run_id: UUID,
            parent_run_id: Optional[UUID] = None,
            **kwargs: Any,
    ) -> Any:
        pass


import time
from enum import Enum
from typing import List, Callable, Optional, Type, Any, Sequence, Union, Dict

import pydantic
from langchain.chains import LLMChain
from langchain.output_parsers.ernie_functions import PydanticOutputFunctionsParser
from langchain.output_parsers.openai_functions import JsonOutputFunctionsParser
from langchain_community.chat_models import ChatOpenAI
from langchain_community.embeddings import OpenAIEmbeddings
from langchain_core.language_models import BaseLanguageModel
from langchain_core.messages import BaseMessage, HumanMessage
from langchain_core.output_parsers import BaseLLMOutputParser
from langchain_core.outputs import Generation
from langchain_core.prompts import BasePromptTemplate, ChatPromptTemplate
from pydantic import BaseModel, Field, create_model
from pydantic.fields import Undefined

from lemon_rag.lemon_runtime import models
from lemon_rag.utils import log


class EmbeddingModel(str, Enum):
    DEFAULT = "text-embedding-ada-002"


Embedding_F = Callable[[List[str]], List[List[float]]]


def get_embedding_function(metrics_name: str) -> Embedding_F:
    embeddings: List[Optional[OpenAIEmbeddings]] = [None]

    def init_embeddings_client():
        if embeddings[0]:
            return
        openai_config: Optional[models.OPENAIConfig] = models.OPENAIConfig.get_or_none()
        if not openai_config:
            raise ValueError("openai_config not found")
        embeddings[0] = OpenAIEmbeddings(openai_api_base=openai_config.base_url, openai_api_key=openai_config.api_key)

    def inner(text: List[str]) -> List[List[float]]:
        init_embeddings_client()
        start_time = time.time()
        client = embeddings[0]
        time_cost = time.time() - start_time
        embedding_key_len = [len(k) for k in text]
        log.info("[Embedding] /%s %.3fs embedding_key_len: %s", metrics_name, time_cost, embedding_key_len)
        return client.embed_documents(text)

    return inner


llm: List[Optional[ChatOpenAI]] = [None]


def get_llm():
    if llm[0] is None:
        openai_config: Optional[models.OPENAIConfig] = models.OPENAIConfig.get_or_none()
        llm[0] = ChatOpenAI(openai_api_base=openai_config.base_url, openai_api_key=openai_config.api_key)
    return llm[0]


class OpenAIFunctionParser(PydanticOutputFunctionsParser):
    def parse_result(self, result: List[Generation], *, partial: bool = True) -> Any:
        if result[0].text:
            return result[0].text
        return super().parse_result(result)


def _get_openai_output_parser(
        functions: List[Type[BaseModel]],
        function_names: Sequence[str],
) -> BaseLLMOutputParser:
    """Get the appropriate function output parser given the user functions."""
    if isinstance(functions[0], type) and issubclass(functions[0], BaseModel):
        if len(functions) > 1:
            pydantic_schema: Union[Dict, Type[BaseModel]] = {
                name: fn for name, fn in zip(function_names, functions)
            }
        else:
            pydantic_schema = functions[0]
        output_parser: BaseLLMOutputParser = OpenAIFunctionParser(
            pydantic_schema=pydantic_schema
        )
    else:
        output_parser = JsonOutputFunctionsParser(args_only=len(functions) <= 1)
    return output_parser


def convert_to_openai_function(function: Type[BaseModel]):
    fields_included = {
        field.name: (field.type_, field.field_info) for field in function.__fields__.values() if
        (not field.field_info.exclude) and (field.field_info.default == Undefined)
    }

    # 动态创建新的类
    new_class = create_model(function.__class__.__name__, **fields_included)
    schema = new_class.schema()
    schema.pop("title")
    return {
        "name": function.__name__,
        "description": function.schema().get("description", ""),
        "parameters": schema,
    }


def create_openai_fn_chain(
        functions: List[Type[BaseModel]],
        llm: BaseLanguageModel,
        prompt: BasePromptTemplate,
        *,
        output_parser: Optional[BaseLLMOutputParser] = None,
        **kwargs: Any,
) -> LLMChain:
    functions = functions or []
    openai_functions = [convert_to_openai_function(f) for f in functions]
    fn_names = [oai_fn["name"] for oai_fn in openai_functions]
    output_parser = output_parser or _get_openai_output_parser(functions, fn_names)
    if openai_functions:
        kwargs.update({
            "llm_kwargs": {"functions": openai_functions},
            "output_key": "function"
        })
    llm_chain = LLMChain(
        llm=llm,
        prompt=prompt,
        output_parser=output_parser,
        **kwargs,
    )
    return llm_chain


class ChatChain(BaseModel):
    """
    一次call的版本信息，messages是浅拷贝的，可以用于重试的时候。
    """
    template: ChatPromptTemplate
    llm: ChatOpenAI = Field(default_factory=get_llm)
    functions: List[Type[BaseModel]] = []

    def run(self, **kwargs):
        if not self.functions:
            print(kwargs, self.template.format_messages(**kwargs))
            return self.llm.generate([self.template.format_messages(**kwargs)])
        chain = create_openai_fn_chain(self.functions, self.llm, self.template)
        return chain.run(**kwargs)




import base64
import os
import subprocess
import tempfile

from pydantic import BaseModel

from lemon_rag.api.base import handle_request_with_pydantic, add_route, handle_chat_auth
from lemon_rag.utils.response_utils import response


class UpgradeRequest(BaseModel):
    package_content: str
    filename: str


@handle_request_with_pydantic(UpgradeRequest)
def handle_upgrade(req: UpgradeRequest):
    with tempfile.TemporaryDirectory() as tmpdirname:
        whl_path = os.path.join(tmpdirname, req.filename)
        with open(whl_path, "wb") as f:
            f.write(base64.b64decode(req.package_content))
        output = subprocess.check_output(
            ['pip', 'install', whl_path],
            stderr=subprocess.STDOUT, universal_newlines=True
        )
    return response(data=output)





system_message_template = """
# Role
You are a awesome assistant. Your responsibility is to assist the user in managing their renovation projects and keeping track of expenses.
* If the user want to record something, you need to call the property function to do that.
* If the user want to query some complicated data, you need to call the function to do that.
* If the user want to ask something about the standard, you need to ask the question based on the given references. When you answer a question, the markdown format is preferred.

You are available to use some tools which is offered as function calls. You can call the function to satisfy the user's requirement if necessary.

Here are some reference data:
{reference_paragraphs}

If you used any reference, you need to output the paragraph code that you used at the last of your answer with a line of json:
{{"referenced_paragraphs": [""]}}

Now it is {time}, Now please start to assistant the user.
"""




from datetime import datetime
from decimal import Decimal
from typing import Union, TypeVar, Generic, Iterator, Optional, TypedDict
import peewee

CharField = Union[peewee.CharField, str]
DatetimeField = Union[peewee.DateTimeField, datetime]
TextField = Union[peewee.TextField, str]
IntegerField = Union[peewee.IntegerField, int]
BooleanField = Union[peewee.BooleanField, bool]
FloatField = Union[peewee.FloatField, float]
DoubleField = Union[peewee.DoubleField, float]
DateField = Union[peewee.DateField, str]
DateTimeField = Union[peewee.DateTimeField, str]
TimeField = Union[peewee.TimeField, str]
DecimalField = Union[peewee.DecimalField, Decimal]
PrimaryKeyField = Union[peewee.PrimaryKeyField, int]

T = TypeVar('T')


class ModelSelect(peewee.ModelSelect, Generic[T]):
    def __iter__(self) -> Iterator[T]:
        pass

    def where(self, *expressions) -> 'ModelSelect[T]':
        pass

    def limit(self, value: Optional[int] = None) -> 'ModelSelect[T]':
        pass

    def offset(self, value: Optional[int] = None) -> 'ModelSelect[T]':
        pass


class BackrefAccessor(peewee.BackrefAccessor, Generic[T]):
    pass


class ModelUpdate(peewee.ModelUpdate, Generic[T]):
    def where(self, *expressions) -> 'ModelUpdate[T]':
        pass

    def execute(self, database=None) -> int:
        pass


class BaseModel(peewee.Model):
    id: PrimaryKeyField


class XFConfig(BaseModel):
    class __InnerFields(TypedDict):
        pass

    @classmethod
    def select(cls, *fields) -> ModelSelect['XFConfig']:
        pass

    @classmethod
    def update(cls, __data=..., **update: __InnerFields) -> ModelUpdate['XFConfig']:
        pass

    @classmethod
    def create(cls, **query: __InnerFields) -> 'XFConfig':
        pass


class OPENAIConfig(BaseModel):
    api_key: CharField
    base_url: CharField

    class __InnerFields(TypedDict):
        api_key: str
        base_url: str

    @classmethod
    def select(cls, *fields) -> ModelSelect['OPENAIConfig']:
        pass

    @classmethod
    def update(cls, __data=..., **update: __InnerFields) -> ModelUpdate['OPENAIConfig']:
        pass

    @classmethod
    def create(cls, **query: __InnerFields) -> 'OPENAIConfig':
        pass


class VectorStoreConfig(BaseModel):
    uri: CharField
    token: CharField

    class __InnerFields(TypedDict):
        uri: str
        token: str

    @classmethod
    def select(cls, *fields) -> ModelSelect['VectorStoreConfig']:
        pass

    @classmethod
    def update(cls, __data=..., **update: __InnerFields) -> ModelUpdate['VectorStoreConfig']:
        pass

    @classmethod
    def create(cls, **query: __InnerFields) -> 'VectorStoreConfig':
        pass


class AuthUserTab(BaseModel):
    username: CharField
    password: CharField
    last_signin: IntegerField
    nickname: CharField
    avatar: CharField
    create_time: IntegerField
    user_quota: Union[BackrefAccessor['UserQuotaTab'], ModelSelect['UserQuotaTab']]
    file: Union[BackrefAccessor['FileUploadRecordTab'], ModelSelect['FileUploadRecordTab']]
    sessions: Union[BackrefAccessor['SessionTab'], ModelSelect['SessionTab']]
    back_sessions: Union[BackrefAccessor['SessionTab'], ModelSelect['SessionTab']]
    messages: Union[BackrefAccessor['MessageTab'], ModelSelect['MessageTab']]
    owned_knowledge_bases: Union[BackrefAccessor['KnowledgeBaseTab'], ModelSelect['KnowledgeBaseTab']]
    tokens: Union[BackrefAccessor['AppAuthTokenTab'], ModelSelect['AppAuthTokenTab']]
    available_knowledge_accesses: Union[
        BackrefAccessor['KnowledgeBaseAccessTab'], ModelSelect['KnowledgeBaseAccessTab']]
    created_accesses: Union[BackrefAccessor['KnowledgeBaseAccessTab'], ModelSelect['KnowledgeBaseAccessTab']]

    class __InnerFields(TypedDict):
        username: str
        password: str
        last_signin: int
        nickname: str
        avatar: str
        create_time: int

    @classmethod
    def select(cls, *fields) -> ModelSelect['AuthUserTab']:
        pass

    @classmethod
    def update(cls, __data=..., **update: __InnerFields) -> ModelUpdate['AuthUserTab']:
        pass

    @classmethod
    def create(cls, **query: __InnerFields) -> 'AuthUserTab':
        pass


class UserQuotaTab(BaseModel):
    request_per_min: IntegerField
    request_per_day: IntegerField
    request_per_month: IntegerField
    max_files: IntegerField
    max_single_file_size: IntegerField
    max_total_knowledge_size: IntegerField
    auth_user: Union[peewee.ForeignKeyField, 'AuthUserTab']

    class __InnerFields(TypedDict):
        request_per_min: int
        request_per_day: int
        request_per_month: int
        max_files: int
        max_single_file_size: int
        max_total_knowledge_size: int

    @classmethod
    def select(cls, *fields) -> ModelSelect['UserQuotaTab']:
        pass

    @classmethod
    def update(cls, __data=..., **update: __InnerFields) -> ModelUpdate['UserQuotaTab']:
        pass

    @classmethod
    def create(cls, **query: __InnerFields) -> 'UserQuotaTab':
        pass


class KnowledgeBaseFileTab(BaseModel):
    filename: CharField
    extension: CharField
    origin_filename: CharField
    file_size: IntegerField
    total_parts: IntegerField
    vectorized_parts: IntegerField
    content_hash: CharField
    knowledge_base: Union[peewee.ForeignKeyField, 'KnowledgeBaseTab']
    upload_records: Union[BackrefAccessor['FileUploadRecordTab'], ModelSelect['FileUploadRecordTab']]
    paragraph_list: Union[BackrefAccessor['KnowledgeParagraph'], ModelSelect['KnowledgeParagraph']]
    sentences: Union[BackrefAccessor['KnowledgeSentence'], ModelSelect['KnowledgeSentence']]

    class __InnerFields(TypedDict):
        filename: str
        extension: str
        origin_filename: str
        file_size: int
        total_parts: int
        vectorized_parts: int
        content_hash: str

    @classmethod
    def select(cls, *fields) -> ModelSelect['KnowledgeBaseFileTab']:
        pass

    @classmethod
    def update(cls, __data=..., **update: __InnerFields) -> ModelUpdate['KnowledgeBaseFileTab']:
        pass

    @classmethod
    def create(cls, **query: __InnerFields) -> 'KnowledgeBaseFileTab':
        pass


class FileUploadRecordTab(BaseModel):
    upload_time: DatetimeField
    first_upload: BooleanField
    file: Union[peewee.ForeignKeyField, 'KnowledgeBaseFileTab']
    uploader: Union[peewee.ForeignKeyField, 'AuthUserTab']

    class __InnerFields(TypedDict):
        upload_time: datetime
        first_upload: bool

    @classmethod
    def select(cls, *fields) -> ModelSelect['FileUploadRecordTab']:
        pass

    @classmethod
    def update(cls, __data=..., **update: __InnerFields) -> ModelUpdate['FileUploadRecordTab']:
        pass

    @classmethod
    def create(cls, **query: __InnerFields) -> 'FileUploadRecordTab':
        pass


class SessionTab(BaseModel):
    create_at: IntegerField
    topic: CharField
    last_msg_id: IntegerField
    last_msg_ts: IntegerField
    assistant_role: CharField
    title: CharField
    disabled_at: BooleanField
    version: IntegerField
    user: Union[peewee.ForeignKeyField, 'AuthUserTab']
    friend: Union[peewee.ForeignKeyField, 'AuthUserTab']
    messages: Union[BackrefAccessor['MessageTab'], ModelSelect['MessageTab']]
    MessageSummarizationTab: Union[BackrefAccessor['MessageSummarizationTab'], ModelSelect['MessageSummarizationTab']]
    sync_history: Union[BackrefAccessor['SyncHistoryTab'], ModelSelect['SyncHistoryTab']]

    class __InnerFields(TypedDict):
        create_at: int
        topic: str
        last_msg_id: int
        last_msg_ts: int
        assistant_role: str
        title: str
        disabled_at: bool
        version: int

    @classmethod
    def select(cls, *fields) -> ModelSelect['SessionTab']:
        pass

    @classmethod
    def update(cls, __data=..., **update: __InnerFields) -> ModelUpdate['SessionTab']:
        pass

    @classmethod
    def create(cls, **query: __InnerFields) -> 'SessionTab':
        pass


class MessageTab(BaseModel):
    msg_id: IntegerField
    role: CharField
    is_system: CharField
    client_ts: IntegerField
    server_ts: IntegerField
    is_answer: BooleanField
    is_system_msg: BooleanField
    support_retry: BooleanField
    support_rating: BooleanField
    content: CharField
    text: CharField
    context_text: CharField
    version: IntegerField
    session: Union[peewee.ForeignKeyField, 'SessionTab']
    user: Union[peewee.ForeignKeyField, 'AuthUserTab']
    MessageSummarizationTab: Union[peewee.ForeignKeyField, 'MessageSummarizationTab']

    class __InnerFields(TypedDict):
        msg_id: int
        role: str
        is_system: str
        client_ts: int
        server_ts: int
        is_answer: bool
        is_system_msg: bool
        support_retry: bool
        support_rating: bool
        content: str
        text: str
        context_text: str
        version: int

    @classmethod
    def select(cls, *fields) -> ModelSelect['MessageTab']:
        pass

    @classmethod
    def update(cls, __data=..., **update: __InnerFields) -> ModelUpdate['MessageTab']:
        pass

    @classmethod
    def create(cls, **query: __InnerFields) -> 'MessageTab':
        pass


class KnowledgeBaseTab(BaseModel):
    name: CharField
    max_files: IntegerField
    create_at: IntegerField
    owner: Union[peewee.ForeignKeyField, 'AuthUserTab']
    files: Union[BackrefAccessor['KnowledgeBaseFileTab'], ModelSelect['KnowledgeBaseFileTab']]
    access_permissions: Union[BackrefAccessor['KnowledgeBaseAccessTab'], ModelSelect['KnowledgeBaseAccessTab']]

    class __InnerFields(TypedDict):
        name: str
        max_files: int
        create_at: int

    @classmethod
    def select(cls, *fields) -> ModelSelect['KnowledgeBaseTab']:
        pass

    @classmethod
    def update(cls, __data=..., **update: __InnerFields) -> ModelUpdate['KnowledgeBaseTab']:
        pass

    @classmethod
    def create(cls, **query: __InnerFields) -> 'KnowledgeBaseTab':
        pass


class DevModuleTab(BaseModel):
    name: CharField
    version: CharField
    registry: CharField

    class __InnerFields(TypedDict):
        name: str
        version: str
        registry: str

    @classmethod
    def select(cls, *fields) -> ModelSelect['DevModuleTab']:
        pass

    @classmethod
    def update(cls, __data=..., **update: __InnerFields) -> ModelUpdate['DevModuleTab']:
        pass

    @classmethod
    def create(cls, **query: __InnerFields) -> 'DevModuleTab':
        pass


class AppAuthTokenTab(BaseModel):
    create_at: IntegerField
    expire_at: IntegerField
    token: CharField
    comment: CharField
    user: Union[peewee.ForeignKeyField, 'AuthUserTab']

    class __InnerFields(TypedDict):
        create_at: int
        expire_at: int
        token: str
        comment: str

    @classmethod
    def select(cls, *fields) -> ModelSelect['AppAuthTokenTab']:
        pass

    @classmethod
    def update(cls, __data=..., **update: __InnerFields) -> ModelUpdate['AppAuthTokenTab']:
        pass

    @classmethod
    def create(cls, **query: __InnerFields) -> 'AppAuthTokenTab':
        pass


class MessageSummarizationTab(BaseModel):
    from_msg_id: IntegerField
    to_msg_id: IntegerField
    summarization: CharField
    msg_count: IntegerField
    SessionTab: Union[peewee.ForeignKeyField, 'SessionTab']
    MessageTab: Union[BackrefAccessor['MessageTab'], ModelSelect['MessageTab']]

    class __InnerFields(TypedDict):
        from_msg_id: int
        to_msg_id: int
        summarization: str
        msg_count: int

    @classmethod
    def select(cls, *fields) -> ModelSelect['MessageSummarizationTab']:
        pass

    @classmethod
    def update(cls, __data=..., **update: __InnerFields) -> ModelUpdate['MessageSummarizationTab']:
        pass

    @classmethod
    def create(cls, **query: __InnerFields) -> 'MessageSummarizationTab':
        pass


class SyncHistoryTab(BaseModel):
    last_read_id: IntegerField
    last_read_ts: IntegerField
    session: Union[peewee.ForeignKeyField, 'SessionTab']

    class __InnerFields(TypedDict):
        last_read_id: int
        last_read_ts: int

    @classmethod
    def select(cls, *fields) -> ModelSelect['SyncHistoryTab']:
        pass

    @classmethod
    def update(cls, __data=..., **update: __InnerFields) -> ModelUpdate['SyncHistoryTab']:
        pass

    @classmethod
    def create(cls, **query: __InnerFields) -> 'SyncHistoryTab':
        pass


class KnowledgeBaseAccessTab(BaseModel):
    permission: IntegerField
    create_at: IntegerField
    knowledge_base: Union[peewee.ForeignKeyField, 'KnowledgeBaseTab']
    user: Union[peewee.ForeignKeyField, 'AuthUserTab']
    creator: Union[peewee.ForeignKeyField, 'AuthUserTab']

    class __InnerFields(TypedDict):
        permission: int
        create_at: int

    @classmethod
    def select(cls, *fields) -> ModelSelect['KnowledgeBaseAccessTab']:
        pass

    @classmethod
    def update(cls, __data=..., **update: __InnerFields) -> ModelUpdate['KnowledgeBaseAccessTab']:
        pass

    @classmethod
    def create(cls, **query: __InnerFields) -> 'KnowledgeBaseAccessTab':
        pass


class KnowledgeParagraph(BaseModel):
    raw_content: CharField
    index: IntegerField
    context_content: CharField
    total_sentences: IntegerField
    file: Union[peewee.ForeignKeyField, 'KnowledgeBaseFileTab']
    sentences: Union[BackrefAccessor['KnowledgeSentence'], ModelSelect['KnowledgeSentence']]

    class __InnerFields(TypedDict):
        raw_content: str
        index: int
        context_content: str
        total_sentences: int

    @classmethod
    def select(cls, *fields) -> ModelSelect['KnowledgeParagraph']:
        pass

    @classmethod
    def update(cls, __data=..., **update: __InnerFields) -> ModelUpdate['KnowledgeParagraph']:
        pass

    @classmethod
    def create(cls, **query: __InnerFields) -> 'KnowledgeParagraph':
        pass


class KnowledgeSentence(BaseModel):
    index: IntegerField
    raw_content: CharField
    context_content: CharField
    vectorized: BooleanField
    paragraph: Union[peewee.ForeignKeyField, 'KnowledgeParagraph']
    file: Union[peewee.ForeignKeyField, 'KnowledgeBaseFileTab']

    class __InnerFields(TypedDict):
        index: int
        raw_content: str
        context_content: str
        vectorized: bool

    @classmethod
    def select(cls, *fields) -> ModelSelect['KnowledgeSentence']:
        pass

    @classmethod
    def update(cls, __data=..., **update: __InnerFields) -> ModelUpdate['KnowledgeSentence']:
        pass

    @classmethod
    def create(cls, **query: __InnerFields) -> 'KnowledgeSentence':
        pass


class TransactionTab(BaseModel):
    voucher_number: CharField
    type: CharField
    item: CharField
    description: CharField
    amount: DecimalField
    notes: CharField
    date: CharField
    timestamp: IntegerField
    create_date: CharField
    create_at: IntegerField
    account: Union[peewee.ForeignKeyField, 'AccountTab']
    TransactionItemTab: Union[peewee.ForeignKeyField, 'TransactionItemTab']
    pre_transaction: Union[BackrefAccessor['PreTransactionTab'], ModelSelect['PreTransactionTab']]

    class __InnerFields(TypedDict):
        voucher_number: str
        type: str
        item: str
        description: str
        amount: Decimal
        notes: str
        date: str
        timestamp: int
        create_date: str
        create_at: int

    @classmethod
    def select(cls, *fields) -> ModelSelect['TransactionTab']:
        pass

    @classmethod
    def update(cls, __data=..., **update: __InnerFields) -> ModelUpdate['TransactionTab']:
        pass

    @classmethod
    def create(cls, **query: __InnerFields) -> 'TransactionTab':
        pass


class AccountTab(BaseModel):
    name: CharField
    amount: DecimalField
    create_date: CharField
    create_at: IntegerField
    transaction: Union[BackrefAccessor['TransactionTab'], ModelSelect['TransactionTab']]
    transfer_in_records: Union[BackrefAccessor['InternalTransferTab'], ModelSelect['InternalTransferTab']]
    transfer_out_records: Union[BackrefAccessor['InternalTransferTab'], ModelSelect['InternalTransferTab']]

    class __InnerFields(TypedDict):
        name: str
        amount: Decimal
        create_date: str
        create_at: int

    @classmethod
    def select(cls, *fields) -> ModelSelect['AccountTab']:
        pass

    @classmethod
    def update(cls, __data=..., **update: __InnerFields) -> ModelUpdate['AccountTab']:
        pass

    @classmethod
    def create(cls, **query: __InnerFields) -> 'AccountTab':
        pass


class InternalTransferTab(BaseModel):
    amount: DecimalField
    description: CharField
    date: CharField
    timestamp: IntegerField
    create_date: CharField
    create_at: IntegerField
    target_account: Union[peewee.ForeignKeyField, 'AccountTab']
    source_account: Union[peewee.ForeignKeyField, 'AccountTab']

    class __InnerFields(TypedDict):
        amount: Decimal
        description: str
        date: str
        timestamp: int
        create_date: str
        create_at: int

    @classmethod
    def select(cls, *fields) -> ModelSelect['InternalTransferTab']:
        pass

    @classmethod
    def update(cls, __data=..., **update: __InnerFields) -> ModelUpdate['InternalTransferTab']:
        pass

    @classmethod
    def create(cls, **query: __InnerFields) -> 'InternalTransferTab':
        pass


class ReceivablePayableTab(BaseModel):
    description: CharField
    type: CharField
    handler: CharField
    product_name: CharField
    specification: CharField
    quantity: DecimalField
    unit_price: DecimalField
    nodes: CharField
    date: CharField
    timestamp: IntegerField
    create_date: CharField
    create_at: IntegerField
    contract: Union[peewee.ForeignKeyField, 'ContractTab']
    pre_transaction: Union[BackrefAccessor['PreTransactionTab'], ModelSelect['PreTransactionTab']]

    class __InnerFields(TypedDict):
        description: str
        type: str
        handler: str
        product_name: str
        specification: str
        quantity: Decimal
        unit_price: Decimal
        nodes: str
        date: str
        timestamp: int
        create_date: str
        create_at: int

    @classmethod
    def select(cls, *fields) -> ModelSelect['ReceivablePayableTab']:
        pass

    @classmethod
    def update(cls, __data=..., **update: __InnerFields) -> ModelUpdate['ReceivablePayableTab']:
        pass

    @classmethod
    def create(cls, **query: __InnerFields) -> 'ReceivablePayableTab':
        pass


class RepaymentItemTab(BaseModel):
    name: CharField
    amount: DecimalField
    create_date: CharField
    create_at: IntegerField
    employee_loan: Union[peewee.ForeignKeyField, 'EmployeeLoanTab']

    class __InnerFields(TypedDict):
        name: str
        amount: Decimal
        create_date: str
        create_at: int

    @classmethod
    def select(cls, *fields) -> ModelSelect['RepaymentItemTab']:
        pass

    @classmethod
    def update(cls, __data=..., **update: __InnerFields) -> ModelUpdate['RepaymentItemTab']:
        pass

    @classmethod
    def create(cls, **query: __InnerFields) -> 'RepaymentItemTab':
        pass


class EmployeeLoanTab(BaseModel):
    description: CharField
    amount: DecimalField
    adjustment_amount: DecimalField
    date: CharField
    timestamp: IntegerField
    create_date: CharField
    create_at: IntegerField
    employee: Union[peewee.ForeignKeyField, 'EmployeeTab']
    repayment_item: Union[BackrefAccessor['RepaymentItemTab'], ModelSelect['RepaymentItemTab']]

    class __InnerFields(TypedDict):
        description: str
        amount: Decimal
        adjustment_amount: Decimal
        date: str
        timestamp: int
        create_date: str
        create_at: int

    @classmethod
    def select(cls, *fields) -> ModelSelect['EmployeeLoanTab']:
        pass

    @classmethod
    def update(cls, __data=..., **update: __InnerFields) -> ModelUpdate['EmployeeLoanTab']:
        pass

    @classmethod
    def create(cls, **query: __InnerFields) -> 'EmployeeLoanTab':
        pass


class EmployeeTab(BaseModel):
    name: CharField
    age: IntegerField
    gender: CharField
    create_date: CharField
    create_at: IntegerField
    project: Union[peewee.ForeignKeyField, 'ProjectTab']
    salary: Union[BackrefAccessor['SalaryTab'], ModelSelect['SalaryTab']]
    employee_loan: Union[BackrefAccessor['EmployeeLoanTab'], ModelSelect['EmployeeLoanTab']]

    class __InnerFields(TypedDict):
        name: str
        age: int
        gender: str
        create_date: str
        create_at: int

    @classmethod
    def select(cls, *fields) -> ModelSelect['EmployeeTab']:
        pass

    @classmethod
    def update(cls, __data=..., **update: __InnerFields) -> ModelUpdate['EmployeeTab']:
        pass

    @classmethod
    def create(cls, **query: __InnerFields) -> 'EmployeeTab':
        pass


class SupplierTab(BaseModel):
    name: CharField
    contact_name: CharField
    contact_phone: CharField
    create_date: CharField
    create_at: IntegerField
    contract: Union[BackrefAccessor['ContractTab'], ModelSelect['ContractTab']]

    class __InnerFields(TypedDict):
        name: str
        contact_name: str
        contact_phone: str
        create_date: str
        create_at: int

    @classmethod
    def select(cls, *fields) -> ModelSelect['SupplierTab']:
        pass

    @classmethod
    def update(cls, __data=..., **update: __InnerFields) -> ModelUpdate['SupplierTab']:
        pass

    @classmethod
    def create(cls, **query: __InnerFields) -> 'SupplierTab':
        pass


class CustomerTab(BaseModel):
    name: CharField
    contact_name: CharField
    contact_phone: CharField
    create_date: CharField
    create_at: IntegerField
    contract: Union[BackrefAccessor['ContractTab'], ModelSelect['ContractTab']]

    class __InnerFields(TypedDict):
        name: str
        contact_name: str
        contact_phone: str
        create_date: str
        create_at: int

    @classmethod
    def select(cls, *fields) -> ModelSelect['CustomerTab']:
        pass

    @classmethod
    def update(cls, __data=..., **update: __InnerFields) -> ModelUpdate['CustomerTab']:
        pass

    @classmethod
    def create(cls, **query: __InnerFields) -> 'CustomerTab':
        pass


class ProjectTab(BaseModel):
    name: CharField
    start_date: CharField
    end_date: CharField
    create_date: CharField
    create_at: IntegerField
    employee: Union[BackrefAccessor['EmployeeTab'], ModelSelect['EmployeeTab']]
    contract: Union[BackrefAccessor['ContractTab'], ModelSelect['ContractTab']]
    pre_transaction: Union[BackrefAccessor['PreTransactionTab'], ModelSelect['PreTransactionTab']]

    class __InnerFields(TypedDict):
        name: str
        start_date: str
        end_date: str
        create_date: str
        create_at: int

    @classmethod
    def select(cls, *fields) -> ModelSelect['ProjectTab']:
        pass

    @classmethod
    def update(cls, __data=..., **update: __InnerFields) -> ModelUpdate['ProjectTab']:
        pass

    @classmethod
    def create(cls, **query: __InnerFields) -> 'ProjectTab':
        pass


class ContractTab(BaseModel):
    number: CharField
    name: CharField
    type: CharField
    amount: DecimalField
    adjustment_amount: DecimalField
    start_date: CharField
    end_date: CharField
    constact_date: CharField
    create_date: CharField
    create_at: IntegerField
    project: Union[peewee.ForeignKeyField, 'ProjectTab']
    customer: Union[peewee.ForeignKeyField, 'CustomerTab']
    supplier: Union[peewee.ForeignKeyField, 'SupplierTab']
    invoice: Union[BackrefAccessor['InvoiceTab'], ModelSelect['InvoiceTab']]
    receivable_payable: Union[BackrefAccessor['ReceivablePayableTab'], ModelSelect['ReceivablePayableTab']]

    class __InnerFields(TypedDict):
        number: str
        name: str
        type: str
        amount: Decimal
        adjustment_amount: Decimal
        start_date: str
        end_date: str
        constact_date: str
        create_date: str
        create_at: int

    @classmethod
    def select(cls, *fields) -> ModelSelect['ContractTab']:
        pass

    @classmethod
    def update(cls, __data=..., **update: __InnerFields) -> ModelUpdate['ContractTab']:
        pass

    @classmethod
    def create(cls, **query: __InnerFields) -> 'ContractTab':
        pass


class InvoiceTab(BaseModel):
    amount: DecimalField
    tax_rate: DecimalField
    type: CharField
    number: CharField
    code: CharField
    content: CharField
    nodes: CharField
    invoice_date: CharField
    create_date: CharField
    create_at: IntegerField
    constact: Union[peewee.ForeignKeyField, 'ContractTab']

    class __InnerFields(TypedDict):
        amount: Decimal
        tax_rate: Decimal
        type: str
        number: str
        code: str
        content: str
        nodes: str
        invoice_date: str
        create_date: str
        create_at: int

    @classmethod
    def select(cls, *fields) -> ModelSelect['InvoiceTab']:
        pass

    @classmethod
    def update(cls, __data=..., **update: __InnerFields) -> ModelUpdate['InvoiceTab']:
        pass

    @classmethod
    def create(cls, **query: __InnerFields) -> 'InvoiceTab':
        pass


class SalaryTab(BaseModel):
    year: IntegerField
    month: IntegerField
    attendance_days: IntegerField
    basic_salary: DecimalField
    overtime_pay: DecimalField
    attendance_bonus: DecimalField
    subsidy: DecimalField
    bonus: DecimalField
    leave_deduction: DecimalField
    other_deductions: DecimalField
    create_date: CharField
    create_at: IntegerField
    employee: Union[peewee.ForeignKeyField, 'EmployeeTab']

    class __InnerFields(TypedDict):
        year: int
        month: int
        attendance_days: int
        basic_salary: Decimal
        overtime_pay: Decimal
        attendance_bonus: Decimal
        subsidy: Decimal
        bonus: Decimal
        leave_deduction: Decimal
        other_deductions: Decimal
        create_date: str
        create_at: int

    @classmethod
    def select(cls, *fields) -> ModelSelect['SalaryTab']:
        pass

    @classmethod
    def update(cls, __data=..., **update: __InnerFields) -> ModelUpdate['SalaryTab']:
        pass

    @classmethod
    def create(cls, **query: __InnerFields) -> 'SalaryTab':
        pass


class PreTransactionTab(BaseModel):
    receivable_payable: Union[peewee.ForeignKeyField, 'ReceivablePayableTab']
    transaction: Union[peewee.ForeignKeyField, 'TransactionTab']
    project: Union[peewee.ForeignKeyField, 'ProjectTab']

    class __InnerFields(TypedDict):
        pass

    @classmethod
    def select(cls, *fields) -> ModelSelect['PreTransactionTab']:
        pass

    @classmethod
    def update(cls, __data=..., **update: __InnerFields) -> ModelUpdate['PreTransactionTab']:
        pass

    @classmethod
    def create(cls, **query: __InnerFields) -> 'PreTransactionTab':
        pass


class TransactionItemTab(BaseModel):
    name: CharField
    create_date: CharField
    create_at: IntegerField
    TransactionTab: Union[BackrefAccessor['TransactionTab'], ModelSelect['TransactionTab']]

    class __InnerFields(TypedDict):
        name: str
        create_date: str
        create_at: int

    @classmethod
    def select(cls, *fields) -> ModelSelect['TransactionItemTab']:
        pass

    @classmethod
    def update(cls, __data=..., **update: __InnerFields) -> ModelUpdate['TransactionItemTab']:
        pass

    @classmethod
    def create(cls, **query: __InnerFields) -> 'TransactionItemTab':
        pass



from concurrent.futures import ThreadPoolExecutor
from typing import List

from lemon_rag.dependencies.data_access import data_access
from lemon_rag.dependencies.vector_access import vector_access, embed_sentence
from lemon_rag.lemon_runtime import models
from lemon_rag.llm.agents.rag.vectorization.content_extractor import Paragraph

vectorization_pool = ThreadPoolExecutor(max_workers=8)


def parse_and_vectorize_document(file: models.KnowledgeBaseFileTab, paragraphs: List[Paragraph]):
    paragraphs_records = data_access.create_paragraphs(file, paragraphs)
    for paragraph in paragraphs_records:
        for sentence in paragraph.sentences:
            vectors = embed_sentence([sentence.raw_content])
            vector_access.save_sentence(sentence, vectors[0])
        data_access.update_file_vectorized_count(file)


def start_parse_document(file: models.KnowledgeBaseFileTab, paragraphs: List[Paragraph]):
    parse_and_vectorize_document(file, paragraphs)
